import logging
from typing import Any, Dict, Optional, Sequence, cast

from ens import ENS
from eth_abi.codec import ABICodec
from web3 import Web3
from web3._utils.abi import build_default_registry
from web3._utils.empty import empty
from web3._utils.module import attach_modules
from web3.main import get_default_modules as get_default_sync_modules
from web3.net import AsyncNet
from web3.providers import BaseProvider

from aioweb3.exceptions import SyncCallDetectedError
from aioweb3.manager import AsyncRequestManager

from .async_eth import AsyncEth
from .provider import AsyncHTTPProvider

MODULES = {'eth': (AsyncEth,), 'net': (AsyncNet,)}


def get_default_async_modules() -> Dict[str, Sequence[Any]]:
    # TODO: comapre this with the result of get_default_sync_modules
    modules = get_default_sync_modules()
    modules["eth"] = (AsyncEth,)
    modules["net"] = (AsyncNet,)
    # modules["version"] = (AsyncVersion,)
    # TODO: parity.personal, geth.{admin,miner,personal,txpool}, testing
    return modules


class AsyncWeb3(Web3):
    logger = logging.getLogger('aioweb3.AsyncWeb3')
    RequestManager = AsyncRequestManager

    # TODO: check ENS to see if it's doing anything network-related
    def __init__(
        self,
        provider: Optional[BaseProvider] = None,
        # provider: BaseProvider,
        middlewares: Optional[Sequence[Any]] = None,
        modules: Optional[Dict[str, Sequence[Any]]] = None,
        ens: ENS = cast(ENS, empty),
    ) -> None:
        # TODO: debugging
        assert self.RequestManager.__name__ == 'AsyncRequestManager'
        if provider:
            assert provider.__class__.__name__ == 'AsyncHTTPProvider'
        else:
            provider = AsyncHTTPProvider()
        if not middlewares:
            middlewares = []
        self.manager = self.RequestManager(self, provider, middlewares)
        self.codec = ABICodec(build_default_registry())
        if modules:
            self.logger.debug("received modules", modules)
        modules = {**get_default_async_modules(), **(modules or {})}
        attach_modules(self, modules)
        self.ens = ens

    def _make_request(self, *args, **kwargs):
        raise SyncCallDetectedError()



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,  # noqa: F401
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from uuid import UUID

from web3._utils.threads import ThreadWithReturn, spawn  # noqa: F401
from web3.datastructures import NamedElementOnion
from web3.manager import RequestManager
from web3.providers.async_base import AsyncBaseProvider
from web3.types import (
    Middleware,
    MiddlewareOnion,
    RPCEndpoint,  # noqa: F401
    RPCResponse,
)

from aioweb3.provider import AsyncHTTPProvider

from .exceptions import SyncCallDetectedError
from .middleware import (
    abi_middleware,
    async_buffered_gas_estimate_middleware,
    async_gas_price_strategy_middleware,
    attrdict_middleware,
    name_to_address_middleware,
    pythonic_middleware,
    request_parameter_normalizer,
    validation_middleware,
)


class AsyncRequestManager(RequestManager):
    def __init__(
        self,
        web3: 'AsyncWeb3',
        provider: Optional[AsyncBaseProvider] = None,
        middlewares: Optional[Sequence[Tuple[Middleware, str]]] = None,
    ) -> None:
        self.web3 = web3
        self.pending_requests: Dict[UUID, ThreadWithReturn[RPCResponse]] = {}

        if middlewares is None:
            middlewares = self.default_middlewares(web3)
        self.middleware_onion: MiddlewareOnion = NamedElementOnion(middlewares)

        if provider is None:
            self.provider = AsyncHTTPProvider()
        else:
            if not isinstance(provider, (AsyncBaseProvider,)):
                raise SyncCallDetectedError(f"are you sure {provider} is async")
            self.provider = provider

    async def _coro_make_request(
        self,
        method: Union[RPCEndpoint, Callable[..., RPCEndpoint]],
        params: Any,
    ) -> RPCResponse:
        # type ignored b/c request_func is an awaitable in async model
        request_func = await self.provider.request_func(  # type: ignore
            self.web3, self.middleware_onion
        )
        self.logger.debug("Making request. Method: %s", method)
        return await request_func(method, params)

    @staticmethod
    def default_middlewares(web3: 'AsyncWeb3') -> List[Tuple[Middleware, str]]:
        """
        List the default middlewares for the request manager.
        Leaving ens unspecified will prevent the middleware from resolving names.
        """
        return [
            (request_parameter_normalizer, 'request_param_normalizer'),
            (async_gas_price_strategy_middleware, 'gas_price_strategy'),
            # (name_to_address_middleware(web3), 'name_to_address'),  Currently is async
            (attrdict_middleware, 'attrdict'),  # No blocking calls
            (pythonic_middleware, 'pythonic'),  # No blocking calls
            (validation_middleware, 'validation'),
            (abi_middleware, 'abi'),
            (async_buffered_gas_estimate_middleware, 'gas_estimate'),
        ]



from typing import Any, Optional, Type, Union

from eth_typing import Address, ChecksumAddress
from web3 import eth
from web3.types import ENS

from aioweb3.contract import AsyncContract
from aioweb3.exceptions import SyncCallDetectedError


class AsyncEth(eth.AsyncEth):
    defaultContractFactory = AsyncContract

    def contract(
        self,
        address: Optional[Union[Address, ChecksumAddress, ENS]] = None,
        **kwargs: Any,
    ) -> Union[Type[AsyncContract], AsyncContract]:

        if "ContractFactoryClass" in kwargs:
            raise SyncCallDetectedError("only supports AsyncContract for now")
        # ContractFactoryClass = kwargs.pop('ContractFactoryClass', self.defaultContractFactory)
        ContractFactory = AsyncContract.factory(self.web3, **kwargs)

        if address:
            return ContractFactory(address)
        else:
            return ContractFactory



import functools
from typing import TYPE_CHECKING, Any, Callable, Sequence

from web3.middleware.abi import abi_middleware  # noqa: F401
from web3.middleware.attrdict import attrdict_middleware  # noqa: F401
from web3.middleware.buffered_gas_estimate import (  # noqa: F401
    async_buffered_gas_estimate_middleware,
    buffered_gas_estimate_middleware,
)
from web3.middleware.cache import (
    _latest_block_based_cache_middleware as latest_block_based_cache_middleware,
)  # noqa: F401
from web3.middleware.cache import (
    _simple_cache_middleware as simple_cache_middleware,
)
from web3.middleware.cache import (
    _time_based_cache_middleware as time_based_cache_middleware,
)
from web3.middleware.cache import (
    construct_latest_block_based_cache_middleware,
    construct_simple_cache_middleware,
    construct_time_based_cache_middleware,
)
from web3.middleware.exception_handling import (
    construct_exception_handler_middleware,
)  # noqa: F401
from web3.middleware.exception_retry_request import (
    http_retry_request_middleware,
)  # noqa: F401
from web3.middleware.filter import local_filter_middleware  # noqa: F401
from web3.middleware.fixture import (  # noqa: F401
    construct_error_generator_middleware,
    construct_fixture_middleware,
    construct_result_generator_middleware,
)
from web3.middleware.formatting import (
    construct_formatting_middleware,
)  # noqa: F401
from web3.middleware.gas_price_strategy import (  # noqa: F401
    async_gas_price_strategy_middleware,
    gas_price_strategy_middleware,
)
from web3.middleware.geth_poa import geth_poa_middleware  # noqa: F401
from web3.middleware.names import name_to_address_middleware  # noqa: F401
from web3.middleware.normalize_request_parameters import (
    request_parameter_normalizer,
)  # noqa: F401
from web3.middleware.pythonic import pythonic_middleware  # noqa: F401
from web3.middleware.signing import (
    construct_sign_and_send_raw_middleware,
)  # noqa: F401
from web3.middleware.stalecheck import make_stalecheck_middleware  # noqa: F401
from web3.middleware.validation import validation_middleware  # noqa: F401
from web3.types import Middleware, RPCEndpoint, RPCResponse

if TYPE_CHECKING:
    from web3 import Web3  # noqa: F401


async def async_combine_middlewares(
    middlewares: Sequence[Middleware],
    web3: 'Web3',
    provider_request_fn: Callable[[RPCEndpoint, Any], Any],
) -> Callable[..., RPCResponse]:
    """
    Returns a callable function which will call the provider.provider_request
    function wrapped with all of the middlewares.
    """
    accumulator_fn = provider_request_fn
    for middleware in reversed(middlewares):
        accumulator_fn = await construct_middleware(
            middleware, accumulator_fn, web3
        )
    return accumulator_fn


async def construct_middleware(
    middleware: Middleware, fn: Callable[..., RPCResponse], w3: 'Web3'
) -> Callable[[RPCEndpoint, Any], Any]:
    return await middleware(fn, w3)
